import datetime
import os
from random import choice, shuffle
import unittest

import cf


def randomise_test_order(*_args):
    '''Return a random choice from 1 or -1.

    When set as the test loader method for standard (merge)sort comparison
    to order all methods in a test case (see 'sortTestMethodsUsing'), ensures
    they run in a random order, meaning implicit reliance on setup or state,
    i.e. test dependencies, become evident over repeated runs.
    '''
    return choice([1, -1])


# Build the test suite from the tests found in the test files.
test_loader = unittest.TestLoader
# Randomise the order to run the test methods within each test case (module),
# i.e. within each test_<TestCase>, e.g. for all in test_AuxiliaryCoordinate:
test_loader.sortTestMethodsUsing = randomise_test_order

testsuite_setup_0 = unittest.TestSuite()
testsuite_setup_0.addTests(
    test_loader().discover('.', pattern='create_test_files.py')
)

# Build the test suite from the tests found in the test files.
testsuite_setup_1 = unittest.TestSuite()
testsuite_setup_1.addTests(
    test_loader().discover('.', pattern='setup_create_field.py')
)

testsuite = unittest.TestSuite()
all_test_cases = test_loader().discover('.', pattern='test_*.py')
# Randomise the order to run the test cases (modules, i.e. test_<TestCase>)
# TODO: change to a in-built unittest way to specify the above (can't find one
# after much searching, but want to avoid mutating weakly-private attribute).
shuffle(all_test_cases._tests)
testsuite.addTests(all_test_cases)

# Run the test suite's first set-up stage.
def run_test_suite_setup_0(verbosity=2):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    runner.run(testsuite_setup_0)


# Run the test suite's second set-up stage.
def run_test_suite_setup_1(verbosity=2):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    runner.run(testsuite_setup_1)


# Run the test suite.
def run_test_suite(verbosity=2):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    outcome = runner.run(testsuite)
    # Note unittest.TextTestRunner().run() does not set an exit code, so (esp.
    # for CI / GH Actions workflows) we need $? = 1 set if any sub-test fails:
    if not outcome.wasSuccessful():
        exit(1)  # else is zero for sucess as standard


if __name__ == '__main__':
    original_chunksize = cf.CHUNKSIZE()
    print('--------------------')
    print('CF-PYTHON TEST SUITE')
    print('--------------------')
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print('')
    print('Running tests from', os.path.abspath(os.curdir))

    cf.CHUNKSIZE(original_chunksize)

    run_test_suite_setup_0()
    run_test_suite_setup_1()
    run_test_suite()

    cf.CHUNKSIZE(original_chunksize)



import cfdm

from .functions import CF

from . import (AuxiliaryCoordinate,
               CellMethod,
               CellMeasure,
               CoordinateReference,
               DimensionCoordinate,
               DomainAncillary,
               DomainAxis,
               Field,
               FieldAncillary,
               Bounds,
               InteriorRing,
               CoordinateConversion,
               Datum,
               Count,
               List,
               Index,
               NodeCountProperties,
               PartNodeCountProperties,
)

from .data import (Data,
                   GatheredArray,
                   NetCDFArray,
                   RaggedContiguousArray,
                   RaggedIndexedArray,
                   RaggedIndexedContiguousArray,
)


class CFImplementation(cfdm.CFDMImplementation):
    '''TODO

    .. versionadded:: 3.0.0

    '''
# --- End: class


_implementation = CFImplementation(
    cf_version = CF(),

    AuxiliaryCoordinate = AuxiliaryCoordinate,
    CellMeasure         = CellMeasure,
    CellMethod          = CellMethod,
    CoordinateReference = CoordinateReference,
    DimensionCoordinate = DimensionCoordinate,
    DomainAncillary     = DomainAncillary,
    DomainAxis          = DomainAxis,
    Field               = Field,
    FieldAncillary      = FieldAncillary,

    Bounds               = Bounds,
    InteriorRing         = InteriorRing,
    CoordinateConversion = CoordinateConversion,
    Datum                = Datum,

    List                    = List,
    Index                   = Index,
    Count                   = Count,
    NodeCountProperties     = NodeCountProperties,
    PartNodeCountProperties = PartNodeCountProperties,

    Data                         = Data,
    GatheredArray                = GatheredArray,
    NetCDFArray                  = NetCDFArray,
    RaggedContiguousArray        = RaggedContiguousArray,
    RaggedIndexedArray           = RaggedIndexedArray,
    RaggedIndexedContiguousArray = RaggedIndexedContiguousArray,
)

def implementation():
    '''Return a container for the CF data model implementation.

    .. versionadded:: 3.1.0

    .. seealso:: `cf.example_field`, `cf.read`, `cf.write`

    :Returns:

        `CFImplementation`
            A container for the CF data model implementation.

    **Examples:**

    >>> i = cf.implementation()
    >>> i
    <CFDMImplementation: >
    >>> i.classes()
    {'AuxiliaryCoordinate': cf.auxiliarycoordinate.AuxiliaryCoordinate,
     'CellMeasure': cf.cellmeasure.CellMeasure,
     'CellMethod': cf.cellmethod.CellMethod,
     'CoordinateReference': cf.coordinatereference.CoordinateReference,
     'DimensionCoordinate': cf.dimensioncoordinate.DimensionCoordinate,
     'DomainAncillary': cf.domainancillary.DomainAncillary,
     'DomainAxis': cf.domainaxis.DomainAxis,
     'Field': cf.field.Field,
     'FieldAncillary': cf.fieldancillary.FieldAncillary,
     'Bounds': cf.bounds.Bounds,
     'InteriorRing': cf.interiorring.InteriorRing,
     'CoordinateConversion': cf.coordinateconversion.CoordinateConversion,
     'Datum': cf.datum.Datum,
     'Data': cf.data.data.Data,
     'GatheredArray': cf.data.gatheredarray.GatheredArray,
     'NetCDFArray': cf.data.netcdfarray.NetCDFArray,
     'RaggedContiguousArray': cf.data.raggedcontiguousarray.RaggedContiguousArray,
     'RaggedIndexedArray': cf.data.raggedindexedarray.RaggedIndexedArray,
     'RaggedIndexedContiguousArray': cf.data.raggedindexedcontiguousarray.RaggedIndexedContiguousArray,
     'List': cf.list.List,
     'Count': cf.count.Count,
     'Index': cf.index.Index,
     'NodeCountProperties': cf.nodecountproperties.NodeCountProperties,
     'PartNodeCountProperties': cf.partnodecountproperties.PartNodeCountProperties}

    '''
    return _implementation.copy()




import datetime
import os
import time
import unittest

import numpy

import cf

class DomainAncillaryTest(unittest.TestCase):
    def test_DomainAncillary(self):
        f = cf.DomainAncillary()
        self.assertTrue(f.isdomainancillary)

        _ = repr(f)
        _ = str(f)
        _ = f.dump(display=False)


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import os
import unittest

import numpy

import cf

class CoordinateReferenceTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    datum = cf.Datum(parameters={'earth_radius': 6371007})

    # Create a vertical grid mapping coordinate reference
    vconversion = cf.CoordinateConversion(
        parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
        domain_ancillaries={'a'   : 'auxiliarycoordinate0',
                            'b'   : 'auxiliarycoordinate1',
                            'orog': 'domainancillary0'})

    vcr = cf.CoordinateReference(
        coordinates=('coord1',),
        datum=datum,
        coordinate_conversion=vconversion)

    # Create a horizontal grid mapping coordinate reference
    hconversion=cf.CoordinateConversion(
        parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                    'grid_north_pole_latitude': 38.0,
                    'grid_north_pole_longitude': 190.0})

    hcr = cf.CoordinateReference(
        coordinate_conversion=hconversion,
        datum=datum,
        coordinates=['x', 'y', 'lat', 'lon']
    )

    def test_CoordinateReference__repr__str__dump(self):
        f = cf.read(self.filename)[0]

        coordinate_conversion=cf.CoordinateConversion(
            parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
            domain_ancillaries={'a': 'aux0', 'b': 'aux1', 'orog': 'orog'})

        datum=cf.Datum(parameters={'earth_radius': 23423423423.34})

        # Create a vertical grid mapping coordinate reference
        t = cf.CoordinateReference(
            coordinates=('coord1',),
            coordinate_conversion=coordinate_conversion,
            datum=datum)

        _ = repr(t)
        _ = str(t)
        _ = t.dump(display=False)

        self.assertFalse(t.has_bounds())

        _ = repr(datum)
        _ = str(datum)

        _ = repr(coordinate_conversion)
        _ = str(coordinate_conversion)


    def test_CoordinateReference_equals(self):
        f = cf.read(self.filename)[0]

        # Create a vertical grid mapping coordinate reference
        t = cf.CoordinateReference(
            coordinates=('coord1',),
            coordinate_conversion=cf.CoordinateConversion(
                parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
                domain_ancillaries={'a': 'aux0', 'b': 'aux1', 'orog': 'orog'})
        )
        self.assertTrue(t.equals(t, verbose=True))
        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cf.CoordinateReference(
            coordinates=['coord1', 'fred', 'coord3'],
            coordinate_conversion=cf.CoordinateConversion(
                parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                            'grid_north_pole_latitude': 38.0,
                            'grid_north_pole_longitude': 190.0})
        )
        self.assertTrue(t.equals(t, verbose=True))
        self.assertTrue(t.equals(t.copy(), verbose=True))

        datum=cf.Datum(parameters={'earth_radius': 6371007})
        conversion=cf.CoordinateConversion(
            parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                        'grid_north_pole_latitude': 38.0,
                        'grid_north_pole_longitude': 190.0})

        t = cf.CoordinateReference(
            coordinate_conversion=conversion,
            datum=datum,
            coordinates=['x', 'y', 'lat', 'lon']
        )

        self.assertTrue(t.equals(t, verbose=True))
        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cf.CoordinateReference(
               coordinates=['coord1', 'fred', 'coord3'],
               coordinate_conversion=cf.CoordinateConversion(
                   parameters={'grid_mapping_name': 'albers_conical_equal_area',
                               'standard_parallel': [-30, 10],
                               'longitude_of_projection_origin': 34.8,
                               'false_easting': -20000,
                               'false_northing': -30000})
        )
        self.assertTrue(t.equals(t, verbose=True))
        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cf.CoordinateReference(
               coordinates=['coord1', 'fred', 'coord3'],
               coordinate_conversion=cf.CoordinateConversion(
                   parameters={'grid_mapping_name': 'albers_conical_equal_area',
                               'standard_parallel': cf.Data([-30, 10]),
                               'longitude_of_projection_origin': 34.8,
                               'false_easting': -20000,
                               'false_northing': -30000})
        )
        self.assertTrue(t.equals(t, verbose=True))
        self.assertTrue(t.equals(t.copy(), verbose=True))


    def test_CoordinateReference_default_value(self):
        f = cf.read(self.filename)[0]

        self.assertTrue(cf.CoordinateReference.default_value('qwerty') == 0.0)
        self.assertTrue(cf.CoordinateReference.default_value('earth_depth') == 0.0)

        cr = f.construct('standard_name:atmosphere_hybrid_height_coordinate')
        self.assertTrue(cr.default_value('qwerty') == 0.0)
        self.assertTrue(cr.default_value('earth_depth') == 0.0)


    def test_CoordinateReference_canonical_units(self):
        f = cf.read(self.filename)[0]

        self.assertTrue(cf.CoordinateReference.canonical_units('qwerty') is None)
        self.assertTrue(cf.CoordinateReference.canonical_units('earth_radius') == cf.Units('m'))

        cr = f.construct('standard_name:atmosphere_hybrid_height_coordinate')
        self.assertTrue(cr.canonical_units('qwerty') is None)
        self.assertTrue(cr.canonical_units('earth_radius') == cf.Units('m'))


    def test_CoordinateReference_match(self):
        self.assertTrue(self.vcr.match())
        self.assertTrue(self.vcr.match('standard_name:atmosphere_hybrid_height_coordinate'))
        self.assertTrue(self.vcr.match('atmosphere_hybrid_height_coordinate'))
        self.assertTrue(self.vcr.match('atmosphere_hybrid_height_coordinate', 'qwerty'))

        self.assertTrue(self.hcr.match())
        self.assertTrue(self.hcr.match('grid_mapping_name:rotated_latitude_longitude'))
        self.assertTrue(self.hcr.match('rotated_latitude_longitude'))
        self.assertTrue(self.hcr.match('grid_mapping_name:rotated_latitude_longitude', 'qwerty'))


    def test_CoordinateReference_get__getitem__(self):
        self.assertTrue(self.vcr['earth_radius'] == self.datum.get_parameter('earth_radius'))
        self.assertTrue(self.vcr['standard_name'] == self.vconversion.get_parameter('standard_name'))
        self.assertTrue(self.vcr.get('earth_radius') is self.datum.get_parameter('earth_radius'))
        self.assertTrue(self.vcr.get('orog') is None)
        self.assertTrue(self.vcr.get('orog', 'qwerty') == 'qwerty')
        self.assertTrue(self.vcr.get('qwerty') is None)
        self.assertTrue(self.vcr['standard_name'] == self.vconversion.get_parameter('standard_name'))
        with self.assertRaises(Exception):
            _ = self.vcr['orog']

        self.assertTrue(self.hcr['earth_radius'] == self.datum.get_parameter('earth_radius'))
        self.assertTrue(self.hcr['grid_north_pole_latitude'] == self.hconversion.get_parameter('grid_north_pole_latitude'))
        self.assertTrue(self.hcr['grid_mapping_name'] == self.hconversion.get_parameter('grid_mapping_name'))
        self.assertTrue(self.hcr.get('earth_radius') is self.datum.get_parameter('earth_radius'))
        self.assertTrue(self.hcr.get('grid_north_pole_latitude', 'qwerty') is self.hconversion.get_parameter('grid_north_pole_latitude'))
        self.assertTrue(self.hcr.get('qwerty') is None)
        self.assertTrue(self.hcr.get('qwerty', 12) == 12)
        with self.assertRaises(Exception):
            _ = self.hcr['qwerty']


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)





from .properties           import Properties
from .propertiesdata       import (PropertiesData,
                                   Subspace)
from .propertiesdatabounds import PropertiesDataBounds
from .coordinate           import Coordinate



import datetime
import os
import time
import unittest

import cf

class DomainAxisTest(unittest.TestCase):
    def test_DomainAxis__repr__str(self):
        x = cf.DomainAxis(size=56)
        x.nc_set_dimension('tas')

        _ = repr(x)
        _ = str(x)


    def test_DomainAxis(self):
        x = cf.DomainAxis(size=111)
        x.nc_set_dimension('tas')

        self.assertTrue(x.size == 111)
        del x.size
        self.assertTrue(getattr(x, 'size', None) is None)
        x.size = 56
        self.assertTrue(x.size == 56)

        self.assertTrue(x == 56)

        x += 1
        self.assertTrue(x.size == 57)
        x -= 1
        self.assertTrue(x.size == 56)
        y = x + 1
        self.assertTrue(y.size == 57)
        y = x - 1
        self.assertTrue(y.size == 55)
        y = 1 + x
        self.assertTrue(y.size == 57)

        self.assertTrue(int(x) == 56)

        self.assertTrue(x >  1)
        self.assertTrue(x >= 1)
        self.assertTrue(x <  100)
        self.assertTrue(x <= 100)
        self.assertTrue(x != 100)

        _ = hash(x)


# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import cfdm

from .constants import cr_coordinates, cr_canonical_units, cr_default_values
from .functions import RTOL, ATOL, allclose
from .functions import inspect as cf_inspect
from .query     import Query

from . import CoordinateConversion
from . import Datum

from .data.data import Data

from .functions import (_DEPRECATION_ERROR_METHOD,
                        _DEPRECATION_ERROR_ATTRIBUTE)

from .decorators import (_inplace_enabled,
                         _inplace_enabled_define_and_cleanup,
                         _deprecated_kwarg_check)


_units = {}


class CoordinateReference(cfdm.CoordinateReference):
    '''A coordinate reference construct of the CF data model.

    A coordinate reference construct relates the coordinate values of
    the coordinate system to locations in a planetary reference frame.

    The domain of a field construct may contain various coordinate
    systems, each of which is constructed from a subset of the
    dimension and auxiliary coordinate constructs. For example, the
    domain of a four-dimensional field construct may contain
    horizontal (y-x), vertical (z), and temporal (t) coordinate
    systems. There may be more than one of each of these, if there is
    more than one coordinate construct applying to a particular
    spatiotemporal dimension (for example, there could be both
    latitude-longitude and y-x projection coordinate systems). In
    general, a coordinate system may be constructed implicitly from
    any subset of the coordinate constructs, yet a coordinate
    construct does not need to be explicitly or exclusively associated
    with any coordinate system.

    A coordinate system of the field construct can be explicitly
    defined by a coordinate reference construct which relates the
    coordinate values of the coordinate system to locations in a
    planetary reference frame and consists of the following:

    * References to the dimension coordinate and auxiliary coordinate
      constructs that define the coordinate system to which the
      coordinate reference construct applies. Note that the coordinate
      values are not relevant to the coordinate reference construct,
      only their properties.

    ..

    * A definition of a datum specifying the zeroes of the dimension
      and auxiliary coordinate constructs which define the coordinate
      system. The datum may be implied by the metadata of the
      referenced dimension and auxiliary coordinate constructs, or
      explicitly provided.

    ..

    * A coordinate conversion, which defines a formula for converting
      coordinate values taken from the dimension or auxiliary
      coordinate constructs to a different coordinate system. A
      coordinate reference construct relates the coordinate values of
      the field to locations in a planetary reference frame.


    **NetCDF interface**

    The netCDF grid mapping variable name of a coordinate reference
    construct may be accessed with the `nc_set_variable`,
    `nc_get_variable`, `nc_del_variable` and `nc_has_variable`
    methods.

    '''
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._CoordinateConversion = CoordinateConversion
        instance._Datum = Datum
        return instance

    def __getitem__(self, key):
        '''Return a parameter value of the datum or the coordinate conversion.

    x.__getitem__(key) <==> x[key]

    If the same parameter exists in both the datum and the coordinate
    conversion then an exception is raised.

    .. seealso:: `get`, `coordinate_conversion.get_parameter`,
                 `datum.get_parameter`

        '''
        out = []
        try:
            out.append(self.coordinate_conversion.get_parameter(key))
        except ValueError:
            try:
                out.append(self.datum.get_parameter(key))
            except ValueError:
                pass
        # --- End: try

        if len(out) == 1:
            return out[0]

        if not out:
            raise KeyError(
                "No {!r} parameter exists in the coordinate conversion nor "
                "the datum".format(key)
            )

        raise KeyError(
            "{!r} parameter exists in both the coordinate conversion and "
            "the datum".format(key)
        )

    def __hash__(self):
        '''x.__hash__() <==> hash(x)

        '''
#        if self.type == 'formula_terms':
#            raise ValueError("Can't hash a formula_terms %s" %
#                             self.__class__.__name__)

        h = sorted(self.items())  # TODO
        h.append(self.identity())

        return hash(tuple(h))

    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return super().__repr__().replace('<', '<CF ', 1)

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _matching_values(self, value0, value1):
        '''TODO
        '''
        if isinstance(value0, Query):
            return bool(value0.evaluate(value1))  # TODO vectors

        try:
            # re.compile object
            return value0.search(value1)
        except (AttributeError, TypeError):
            return self._equals(value1, value0)

    # ----------------------------------------------------------------
    # Private attributes
    # ----------------------------------------------------------------
    @property
    def _coordinate_identities(self):
        '''TODO

    .. versionadded:: 3.0.0

        '''
        return cr_coordinates.get(self.identity(), ())

    def has_bounds(self):
        '''Returns False since coordinate reference constructs do not have
    cell bounds.

    **Examples:**

    >>> c.has_bounds()
    False

        '''
        return False


#    def canonical(self, field=None):
#        '''
#        '''
#        ref = self.copy()
#
#        for term, value in ref.parameters.iteritems():
#            if value is None or isinstance(value, str):
#                continue
#
#            canonical_units = self.canonical_units(term)
#            if canonical_units is None:
#                continue
#
#            if isinstance(canonical_units, str):
#                # units is a standard_name of a coordinate
#                if field is None:
#                    raise ValueError("Set the field parameter")
#                coord = field.coord(canonical_units, exact=True)
#                if coord is not None:
#                    canonical_units = coord.Units
#
#            if canonical_units is not None:
#                units = getattr(value, 'Units', None)
#                if units is not None:
#                    if not canonical_units.equivalent(units):
#                        raise ValueError("xasdddddddddddddd 87236768 TODO")
#                    value.Units = canonical_units
#        # --- End: for
#
#        return ref

    @classmethod
    def canonical_units(cls, term):
        '''Return the canonical units for a standard CF coordinate conversion
    term.

    :Parameters:

        term: `str`
            The name of the term.

    :Returns:

        `Units` or `None`
            The canonical units, or `None` if there are not any.

    **Examples:**

    >>> cf.CoordinateReference.canonical_units('perspective_point_height')
    <Units: m>
    >>> cf.CoordinateReference.canonical_units('ptop')
    None

        '''
        return cr_canonical_units.get(term, None)

    def close(self):
        '''Close all files referenced by coordinate conversion term values.

    :Returns:

        `None`

    **Examples:**

    >>> c.close()

        '''
        pass

    @classmethod
    def default_value(cls, term):
        '''Return the default value for an unset standard CF coordinate
    conversion term.

    :Parameters:

        term: `str`
            The name of the term.

    :Returns:

            The default value, or 0.0 if one is not available.

    **Examples:**

    >>> cf.CoordinateReference.default_value('ptop')
    0.0
    >>> print(cf.CoordinateReference.default_value('north_pole_grid_latitude'))
    0.0

        '''
        return cr_default_values.get(term, 0.0)

    @_deprecated_kwarg_check('traceback')
    def equivalent(self, other, atol=None, rtol=None, verbose=False,
                   traceback=False):
        '''True if two coordinate references are logically equal, False
    otherwise.

    :Parameters:

        other: cf.CoordinateReference
            The object to compare for equality.

        atol: `float`, optional
            The absolute tolerance for all numerical comparisons, By
            default the value returned by the `cf.ATOL` function is used.

        rtol: `float`, optional
            The relative tolerance for all numerical comparisons, By
            default the value returned by the `cf.RTOL` function is used.

        traceback: `bool`, optional
            If True then print a traceback highlighting where the two
            instances differ. TODO

    :Returns:

        out: `bool`
            Whether or not the two objects are equivalent.

    **Examples:**

    TODO

        '''
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if verbose:
                print("{}: Different types ({!r} != {!r})".format(
                    self.__class__.__name__,
                    self.__class__.__name__,
                    other.__class__.__name__)
                )  # pragma: no cover
            return False

        # ------------------------------------------------------------
        # Check the name
        # ------------------------------------------------------------
        if self.identity() != other.identity():
            if verbose:
                print("{}: Different identities ({!r} != {!r})".format(
                    self.__class__.__name__,
                    self.identity(),
                    other.identity()
                ))  # pragma: no cover
            return False

        # ------------------------------------------------------------
        # Check the domain ancillary terms
        # ------------------------------------------------------------
#        ancillaries0 = self._conversion['ancillary']
#        ancillaries1 = other._conversion['ancillary']
        ancillaries0 = self.coordinate_conversion.domain_ancillaries()
        ancillaries1 = other.coordinate_conversion.domain_ancillaries()
        if set(ancillaries0) != set(ancillaries1):
            if verbose:
                print("{}: Non-equivalent domain ancillary terms".format(
                    self.__class__.__name__)
                )  # pragma: no cover
            return False

        # Check that if one term is None then so is the other
        for term, value0 in ancillaries0.items():
            if (value0 is None) != (ancillaries1[term] is None):
                if verbose:
                    print(
                        "{}: Non-equivalent domain ancillary-valued "
                        "term {!r}".format(self.__class__.__name__,  term)
                    )  # pragma: no cover
                return False
        # --- End: for

        # ------------------------------------------------------------
        # Check the parameter terms and their values
        # ------------------------------------------------------------
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

#        parameters0 = self._conversion['parameter']
#        parameters1 = other._conversion['parameter']
        parameters0 = self.coordinate_conversion.parameters()
        parameters1 = other.coordinate_conversion.parameters()

        for term in set(parameters0).union(parameters1):
            value0 = parameters0.get(term, None)
            value1 = parameters1.get(term, None)

            if value1 is None and value0 is None:
                # Term is unset in self and other
                continue

            if value0 is None:
                # Term is unset in self
                value0 = self.default_value(term)

            if value1 is None:
                # Term is unset in other
                value1 = other.default_value(term)

            if not allclose(value0, value1, rtol=rtol, atol=atol):
                if verbose:
                    print(
                        "{}: Non-equivalent coordinate conversion parameter-"
                        "valued term {!r}".format(
                            self.__class__.__name__, term)
                    )  # pragma: no cover
                return False
        # --- End: for

        parameters0 = self.datum.parameters()
        parameters1 = other.datum.parameters()

        for term in set(parameters0).union(parameters1):
            value0 = parameters0.get(term, None)
            value1 = parameters1.get(term, None)

            if value1 is None and value0 is None:
                # Term is unset in self and other
                continue

            if value0 is None:
                # Term is unset in self
                value0 = self.default_value(term)

            if value1 is None:
                # Term is unset in other
                value1 = other.default_value(term)

            if not allclose(value0, value1, rtol=rtol, atol=atol):
                if verbose:
                    print(
                        "{}: Non-equivalent datum parameter-valued "
                        "term {!r}".format(self.__class__.__name__, term)
                    )  # pragma: no cover
                return False
        # --- End: for

        # Still here?
        return True

    def get(self, key, default=None):
        '''Return a parameter value of the datum or the coordinate conversion.

    .. versionadded:: 3.0.0

    .. seealso:: `coordinate_conversion.get_parameter`,
                 `datum.get_parameter`, `__getitem__`,

    :Parameters:

        key: `str` TODO

        default: optional TODO

    :Returns:

            The parameter value.

        '''
        try:
            return self[key]
        except KeyError:
            return default

    def inspect(self):
        '''Inspect the attributes.

    .. seealso:: `cf.inspect`

    :Returns:

        `None`

        '''
        print(cf_inspect(self))  # pragma: no cover

    def match_by_identity(self, *identities):
        '''Determine whether or not one of the identities matches.

    .. versionadded:: 3.0.0

    .. seealso:: `identities`, `match`

    :Parameters:

        TODO

    :Returns:

        `bool`
            Whether or not the coordinate reference matches one of the
            given identities.

    **Examples:**

    TODO

        '''
        if not identities:
            return True

        self_identities = self.identities()

        x = self.coordinate_conversion.get_parameter('grid_mapping_name', None)
        if x is not None:
            self_identities.insert(0, x)

        x = self.coordinate_conversion.get_parameter('standard_name', None)
        if x is not None:
            self_identities.insert(0, x)

        ok = False
        for value0 in identities:
            for value1 in self_identities:
                ok = self._matching_values(value0, value1)
                if ok:
                    break
            # --- End: for

            if ok:
                break
        # --- End: for

        return ok

    def match(self, *identities):
        '''Alias for `cf.CoordinateReference.match_by_identity`

        '''
        return self.match_by_identity(*identities)

    @_deprecated_kwarg_check('i')
    @_inplace_enabled
    def change_identifiers(self, identity_map, coordinate=True,
                           ancillary=True, strict=False,
                           inplace=False, i=False):
        '''Change the TODO

    If an identifier is not in the provided mapping then it is
    set to `None` and thus effectively removed from the coordinate
    reference.

    :Parameters:

        identity_map: dict
            For example: ``{'dim2': 'dim3', 'aux2': 'latitude', 'aux4': None}``

        strict: `bool`, optional
            If True then coordinate or domain ancillary identifiers not
            set in the *identity_map* dictiontary are set to `None`. By
            default they are left unchanged.

        i: `bool`, optional

    :Returns:

        `None`

    **Examples:**

    >>> r = cf.CoordinateReference('atmosphere_hybrid_height_coordinate',
    ...                             a='ncvar:ak',
    ...                             b='ncvar:bk')
    >>> r.coordinates
    {'atmosphere_hybrid_height_coordinate'}
    >>> r.change_coord_identitiers({
    ...     'atmosphere_hybrid_height_coordinate', 'dim1', 'ncvar:ak': 'aux0'
    ... })
    >>> r.coordinates
    {'dim1', 'aux0'}

        '''
        r = _inplace_enabled_define_and_cleanup(self)

        if not identity_map and not strict:
            if inplace:
                r = None
            return r

        if strict:
            default = None

        if ancillary:
            for term, identifier in (
                    r.coordinate_conversion.domain_ancillaries().items()):
                if not strict:
                    default = identifier
                r.coordinate_conversion.set_domain_ancillary(
                    term, identity_map.get(identifier, default), copy=False
                )
        # --- End: if

        if coordinate:
            for identifier in r.coordinates():
                if not strict:
                    default = identifier

                r.del_coordinate(identifier)
                r.set_coordinate(identity_map.get(identifier, default))
        # --- End: if

        r.del_coordinate(None, None)

        return r

    def creation_commands(self, representative_data=False,
                          namespace='cf', indent=0, string=True,
                          name='c'):
        '''Return the commands that would create the field construct.

    **Construct keys**

    The *key* parameter of the output `set_construct` commands is
    utilised in order minimise the number of commands needed to
    implement cross-referencing between constructs (e.g. between a
    coordinate reference construct and coordinate constructs). This is
    usually not necessary when building field constructs, as by
    default the `set_construct` method returns a unique construct key
    for the construct being set.

    .. versionadded:: 3.2.0

    .. seealso:: `cf.Data.creation_commands`,
                 `cf.Field.creation_commands`

    :Parameters:

        representative_data: `bool`, optional
            Return one-line representations of `Data` instances, which
            are not executable code but prevent the data being
            converted in its entirety to a string representation.

        namespace: `str`, optional
            The namespace containing classes of the ``cf``
            package. This is prefixed to the class name in commands
            that instantiate instances of ``cf`` objects. By default,
            *namespace* is ``'cf'``, i.e. it is assumed that ``cf``
            was imported as ``import cf``.

            *Parameter example:*
              If ``cf`` was imported as ``import cf as cfp`` then set
              ``namespace='cfp'``

            *Parameter example:*
              If ``cf`` was imported as ``from cf import *`` then set
              ``namespace=''``

        indent: `int`, optional
            Indent each line by this many spaces. By default no
            indentation is applied. Ignored if *string* is False.

        string: `bool`, optional
            If False then return each command as an element of a
            `list`. By default the commands are concatenated into
            a string, with a new line inserted between each command.

    :Returns:

        `str` or `list`
            The commands in a string, with a new line inserted between
            each command. If *string* is False then the separate
            commands are returned as each element of a `list`.

    **Examples:**

    >>> q = cf.example_field(0)
    >>> print(q)
    Field: specific_humidity (ncvar%q)
    ----------------------------------
    Data            : specific_humidity(latitude(5), longitude(8)) 1
    Cell methods    : area: mean
    Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                    : longitude(8) = [22.5, ..., 337.5] degrees_east
                    : time(1) = [2019-01-01 00:00:00]
    >>> print(q.creation_commands())
    # field: specific_humidity
    f = cf.Field()
    #
    f.set_properties({'Conventions': 'CF-1.7', 'project': 'research', 'standard_name': 'specific_humidity', 'units': '1'})
    f.nc_set_variable('q')
    #
    # domain_axis
    c = cf.DomainAxis(size=5)
    c.nc_set_dimension('lat')
    f.set_construct(c, key='domainaxis0')
    #
    # domain_axis
    c = cf.DomainAxis(size=8)
    c.nc_set_dimension('lon')
    f.set_construct(c, key='domainaxis1')
    #
    # domain_axis
    c = cf.DomainAxis(size=1)
    f.set_construct(c, key='domainaxis2')
    #
    # field data
    data = cf.Data([[0.007, 0.034, 0.003, 0.014, 0.018, 0.037, 0.024, 0.029], [0.023, 0.036, 0.045, 0.062, 0.046, 0.073, 0.006, 0.066], [0.11, 0.131, 0.124, 0.146, 0.087, 0.103, 0.057, 0.011], [0.029, 0.059, 0.039, 0.07, 0.058, 0.072, 0.009, 0.017], [0.006, 0.036, 0.019, 0.035, 0.018, 0.037, 0.034, 0.013]], units='1', dtype='f8')
    f.set_data(data, axes=('domainaxis0', 'domainaxis1'))
    #
    # dimension_coordinate
    c = cf.DimensionCoordinate()
    c.set_properties({'units': 'degrees_north', 'standard_name': 'latitude'})
    c.nc_set_variable('lat')
    data = cf.Data([-75.0, -45.0, 0.0, 45.0, 75.0], units='degrees_north', dtype='f8')
    c.set_data(data)
    b = cf.Bounds()
    b.set_properties({'units': 'degrees_north'})
    b.nc_set_variable('lat_bnds')
    data = cf.Data([[-90.0, -60.0], [-60.0, -30.0], [-30.0, 30.0], [30.0, 60.0], [60.0, 90.0]], units='degrees_north', dtype='f8')
    b.set_data(data)
    c.set_bounds(b)
    f.set_construct(c, axes=('domainaxis0',), key='dimensioncoordinate0', copy=False)
    #
    # dimension_coordinate
    c = cf.DimensionCoordinate()
    c.set_properties({'units': 'degrees_east', 'standard_name': 'longitude'})
    c.nc_set_variable('lon')
    data = cf.Data([22.5, 67.5, 112.5, 157.5, 202.5, 247.5, 292.5, 337.5], units='degrees_east', dtype='f8')
    c.set_data(data)
    b = cf.Bounds()
    b.set_properties({'units': 'degrees_east'})
    b.nc_set_variable('lon_bnds')
    data = cf.Data([[0.0, 45.0], [45.0, 90.0], [90.0, 135.0], [135.0, 180.0], [180.0, 225.0], [225.0, 270.0], [270.0, 315.0], [315.0, 360.0]], units='degrees_east', dtype='f8')
    b.set_data(data)
    c.set_bounds(b)
    f.set_construct(c, axes=('domainaxis1',), key='dimensioncoordinate1', copy=False)
    #
    # dimension_coordinate
    c = cf.DimensionCoordinate()
    c.set_properties({'units': 'days since 2018-12-01', 'standard_name': 'time'})
    c.nc_set_variable('time')
    data = cf.Data([31.0], units='days since 2018-12-01', dtype='f8')
    c.set_data(data)
    f.set_construct(c, axes=('domainaxis2',), key='dimensioncoordinate2', copy=False)
    #
    # cell_method
    c = cf.CellMethod()
    c.set_method('mean')
    c.set_axes(('area',))
    f.set_construct(c)

    >>> print(q.creation_commands(representative_data=True, namespace='', indent=4))
        # field: specific_humidity
        f = Field()
        #
        f.set_properties({'Conventions': 'CF-1.7', 'project': 'research', 'standard_name': 'specific_humidity', 'units': '1'})
        f.nc_set_variable('q')
        #
        # domain_axis
        c = DomainAxis(size=5)
        c.nc_set_dimension('lat')
        f.set_construct(c, key='domainaxis0')
        #
        # domain_axis
        c = DomainAxis(size=8)
        c.nc_set_dimension('lon')
        f.set_construct(c, key='domainaxis1')
        #
        # domain_axis
        c = DomainAxis(size=1)
        f.set_construct(c, key='domainaxis2')
        #
        # field data
        data = <CF Data(5, 8): [[0.007, ..., 0.013]] 1> # Representative data
        f.set_data(data, axes=('domainaxis0', 'domainaxis1'))
        #
        # dimension_coordinate
        c = DimensionCoordinate()
        c.set_properties({'units': 'degrees_north', 'standard_name': 'latitude'})
        c.nc_set_variable('lat')
        data = <CF Data(5): [-75.0, ..., 75.0] degrees_north> # Representative data
        c.set_data(data)
        b = Bounds()
        b.set_properties({'units': 'degrees_north'})
        b.nc_set_variable('lat_bnds')
        data = <CF Data(5, 2): [[-90.0, ..., 90.0]] degrees_north> # Representative data
        b.set_data(data)
        c.set_bounds(b)
        f.set_construct(c, axes=('domainaxis0',), key='dimensioncoordinate0', copy=False)
        #
        # dimension_coordinate
        c = DimensionCoordinate()
        c.set_properties({'units': 'degrees_east', 'standard_name': 'longitude'})
        c.nc_set_variable('lon')
        data = <CF Data(8): [22.5, ..., 337.5] degrees_east> # Representative data
        c.set_data(data)
        b = Bounds()
        b.set_properties({'units': 'degrees_east'})
        b.nc_set_variable('lon_bnds')
        data = <CF Data(8, 2): [[0.0, ..., 360.0]] degrees_east> # Representative data
        b.set_data(data)
        c.set_bounds(b)
        f.set_construct(c, axes=('domainaxis1',), key='dimensioncoordinate1', copy=False)
        #
        # dimension_coordinate
        c = DimensionCoordinate()
        c.set_properties({'units': 'days since 2018-12-01', 'standard_name': 'time'})
        c.nc_set_variable('time')
        data = <CF Data(1): [2019-01-01 00:00:00]> # Representative data
        c.set_data(data)
        f.set_construct(c, axes=('domainaxis2',), key='dimensioncoordinate2', copy=False)
        #
        # cell_method
        c = CellMethod()
        c.set_method('mean')
        c.set_axes(('area',))
        f.set_construct(c)

        '''
        namespace0 = namespace
        if namespace0:
            namespace = namespace+"."
        else:
            namespace = ""

        indent = ' ' * indent

        out = []
        out.append("# "+self.construct_type)
        out.append("{} = {}{}()".format(name, namespace,
                                        self.__class__.__name__))

        nc = self.nc_get_variable(None)
        if nc is not None:
            out.append("{}.nc_set_variable({!r})".format(name, nc))

        coordinates = self.coordinates()
        if coordinates:
            out.append("{}.set_coordinates({})".format(name,
                                                       coordinates))

        for term, value in self.datum.parameters().items():
            if isinstance(value, Data):
                value = value.creation_commands(name=None,
                                                namespace=namespace0,
                                                indent=0,
                                                string=False)
            else:
                value = repr(value)

            out.append("{}.datum.set_parameter({!r}, {})".format(
                name, term, value))

        for term, value in self.coordinate_conversion.parameters().items():
            if isinstance(value, Data):
                value = value.creation_commands(name=None,
                                                namespace=namespace0,
                                                indent=0,
                                                string=False)
            else:
                value = repr(value)

            out.append("{}.coordinate_conversion.set_parameter({!r}, {})".format(
                name, term, value))

        domain_ancillaries = self.coordinate_conversion.domain_ancillaries()
        if domain_ancillaries:
            out.append("{}.coordinate_conversion.set_domain_ancillaries({})".format(
                name, domain_ancillaries))

        if string:
            out[0] = indent+out[0]
            out = ('\n'+indent).join(out)

        return out

    def structural_signature(self, rtol=None, atol=None):
        '''TODO

    :Return:

        `tuple`
        '''
        s = [self.identity()]
        append = s.append

        for component in ('datum', 'coordinate_conversion'):
            x = getattr(self, component)
            for term, value in sorted(x.parameters().items()):
                if isinstance(value, str):
                    append((component+':'+term, value))
                    continue

                if value is None:
                    # Do not add an unset scalar or vector parameter value
                    # to the structural signature
                    continue

                value = Data.asdata(value, dtype=float)

                cu = self.canonical_units(term)
                if cu is not None:
                    if value.Units.equivalent(cu):
                        value.Units = cu
                    elif value.Units:
                        cu = value.Units
                else:
                    cu = value.Units

                if str(cu) in _units:
                    cu = _units[str(cu)]
                else:
                    ok = 0
                    for units in _units.values():
                        if cu.equals(units):
                            _units[str(cu)] = units
                            cu = units
                            ok = 1
                            break
                    if not ok:
                        _units[str(cu)] = cu

                if allclose(
                        value, self.default_value(term), rtol=rtol, atol=atol):
                    # Do not add a default value to the structural signature
                    continue

                append((
                    component + ':' + term, value,
                    cu.formatted(definition=True)
                ))
        # --- End: for

        # Add the domain ancillary-valued terms which have been set
        terms = self.coordinate_conversion.domain_ancillaries()
        append(tuple(sorted(
            [term for term, value in terms.items() if value is not None]
        )))

        return tuple(s)

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def __delitem__(self, key):
        '''x.__delitem__(key) <==> del x[key]

    Deprecated at version 3.0.0. Use method 'datum.del_parameter',
    'coordinate_conversion.del_parameter' or
    'coordinate_conversion.del_domain_ancillary' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, '__getitem__',
            "Use method 'datum.del_parameter', "
            "'coordinate_conversion.del_parameter' or "
            "'coordinate_conversion.del_domain_ancillary' instead."
        )  # pragma: no cover

    @property
    def conversion(self):
        '''Deprecated at version 3.0.0. Use attribute 'coordinate_conversion'
    instead.

        '''
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, 'conversion',
            "Use attribute 'coordinate_conversion' instead."
        )  # pragma: no cover

    @property
    def hasbounds(self):
        '''False. Coordinate reference objects do not have cell bounds.

    Deprecated at version 3.0.0. Use method 'has_bounds' instead.

        '''
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, 'hasbounds', "Use method 'has_bounds' instead."
        )  # pragma: no cover

    @property
    def ancillaries(self):
        '''Deprecated at version 3.0.0. Use the
    'coordinate_conversion.domain_ancillaries' method instead.

        '''
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, 'ancillaries',
            "Use the 'coordinate_conversion.domain_ancillaries' method "
            "instead."
        )  # pragma: no cover

    @property
    def parameters(self):
        '''Deprecated at version 3.0.0. Use methods
    'coordinate_conversion.parameters' and 'datum.parameters' instead.

        '''
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, 'parameters',
            "Use methods 'coordinate_conversion.parameters' and "
            "'datum.parameters' instead."
        )  # pragma: no cover

    def clear(self, coordinates=True, parameters=True, ancillaries=True):
        '''Deprecated at version 3.0.0. Use methods
    'coordinate_conversion.parameters' and 'datum.parameters' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'parameters',
            "Use methods 'coordinate_conversion.parameters' and "
            "'datum.parameters' instead."
        )  # pragma: no cover

    def name(self, default=None, identity=False, ncvar=False):
        '''Return a name.

    Deprecated at version 3.0.0. Use the 'identity' method instead.
        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'name', "Use the 'identity' method instead."
        )  # pragma: no cover

    def all_identifiers(self):
        '''Deprecated at version 3.0.0.

        '''
        _DEPRECATION_ERROR_METHOD(self, 'all_identifiers')  # pragma: no cover

    def set_term(self, term_type, term, value):
        '''Deprecated at version 3.0.0. Use method 'datum.set_parameter',
    'coordinate_conversion.set_parameter' or
    'coordinate_conversion.set_domain_ancillary' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'set_term',
            "Use method 'datum.set_parameter', "
            "'coordinate_conversion.set_parameter' or "
            "'coordinate_conversion.set_domain_ancillary' instead."
        )  # pragma: no cover

# --- End: class



import os

from glob    import glob
from os.path import isdir

from .netcdf import NetCDFRead
from .um     import UMRead

from ..cfimplementation import implementation

from ..fieldlist import FieldList
from ..functions import flat

from ..aggregate import aggregate as cf_aggregate

from ..functions import _DEPRECATION_ERROR_FUNCTION_KWARGS

# --------------------------------------------------------------------
# Create an implementation container and initialize a read object for
# each format
# --------------------------------------------------------------------
_implementation = implementation()
netcdf = NetCDFRead(_implementation)
UM = UMRead(_implementation)


def read(files, external=None, verbose=False, warnings=False,
         ignore_read_error=False, aggregate=True, nfields=None,
         squeeze=False, unsqueeze=False, fmt=None, select=None,
         extra=None, recursive=False, followlinks=False, um=None,
         chunk=True, field=None, height_at_top_of_model=None,
         select_options=None, follow_symlinks=False):
    '''Read field constructs from netCDF, CDL, PP or UM fields files.

    NetCDF files may be on disk or on an OPeNDAP server.

    Any amount of files of any combination of file types may be read.

    **NetCDF unlimited dimensions**

    Domain axis constructs that correspond to NetCDF unlimited
    dimensions may be accessed with the
    `~cf.DomainAxis.nc_is_unlimited` and
    `~cf.DomainAxis.nc_set_unlimited` methods of a domain axis
    construct.

    **CF-compliance**

    If the dataset is partially CF-compliant to the extent that it is
    not possible to unambiguously map an element of the netCDF dataset
    to an element of the CF data model, then a field construct is
    still returned, but may be incomplete. This is so that datasets
    which are partially conformant may nonetheless be modified in
    memory and written to new datasets.

    Such "structural" non-compliance would occur, for example, if the
    "coordinates" attribute of a CF-netCDF data variable refers to
    another variable that does not exist, or refers to a variable that
    spans a netCDF dimension that does not apply to the data
    variable. Other types of non-compliance are not checked, such
    whether or not controlled vocabularies have been adhered to. The
    structural compliance of the dataset may be checked with the
    `~cf.Field.dataset_compliance` method of the field construct, as
    well as optionally displayed when the dataset is read by setting
    the warnings parameter.

    **CDL files**

    A file is considered to be a CDL representation of a netCDF
    dataset
    (https://www.unidata.ucar.edu/software/netcdf/netcdf/CDL-Syntax.html)
    if it is a text file that starts with the seven characters "netcdf
    " (six letters followed by a space). It is converted to a
    temporary netCDF4 file using the external ``ncgen`` command, and
    the temporary file persists until the end of the Python session,
    at which time it is automatically deleted. The CDL file may omit
    data array values (as would be the case, for example, if the file
    was created with the ``-h`` or ``-c`` option to ``ncdump``), in
    which case the relevant constructs in memory will be created
    with data containing missing values.

    **PP and UM fields files**

    32-bit and 64-bit PP and UM fields files of any endian-ness can be
    read. In nearly all cases the file format is auto-detected from
    the first 64 bits in the file, but for the few occasions when this
    is not possible, the *um* keyword allows the format to be
    specified, as well as the UM version (if the latter is not
    inferrable from the PP or lookup header information).

    2-d "slices" within a single file are always combined, where
    possible, into field constructs with 3-d, 4-d or 5-d data. This is
    done prior to the field construct aggregation carried out by the
    `cf.read` function.

    When reading PP and UM fields files, the *relaxed_units* aggregate
    option is set to `True` by default, because units are not always
    available to field constructs derived from UM fields files or PP
    files.

    **Performance**

    Descriptive properties are always read into memory, but lazy
    loading is employed for all data arrays, which means that no data
    is read into memory until the data is required for inspection or
    to modify the array contents. This maximises the number of field
    constructs that may be read within a session, and makes the read
    operation fast.

    .. seealso:: `cf.aggregate`, `cf.load_stash2standard_name`,
                 `cf.write`, `cf.Field.convert`,
                 `cf.Field.dataset_compliance`

    :Parameters:

        files: (arbitrarily nested sequence of) `str`
            A string or arbitrarily nested sequence of strings giving
            the file names, directory names, or OPenDAP URLs from
            which to read field constructs. Various type of expansion
            are applied to the names:

            ====================  ======================================
            Expansion             Description
            ====================  ======================================
            Tilde                 An initial component of ``~`` or
                                  ``~user`` is replaced by that *user*'s
                                  home directory.

            Environment variable  Substrings of the form ``$name`` or
                                  ``${name}`` are replaced by the value
                                  of environment variable *name*.

            Pathname              A string containing UNIX file name
                                  metacharacters as understood by the
                                  Python `glob` module is replaced by
                                  the list of matching file names. This
                                  type of expansion is ignored for
                                  OPenDAP URLs.
            ====================  ======================================

            Where more than one type of expansion is used in the same
            string, they are applied in the order given in the above
            table.

            *Parameter example:*
              The file ``file.nc`` in the user's home directory could
              be described by any of the following:
              ``'$HOME/file.nc'``, ``'${HOME}/file.nc'``,
              ``'~/file.nc'``, ``'~/tmp/../file.nc'``.

            When a directory is specified, all files in that directory
            are read. Sub-directories are not read unless the
            *recursive* parameter is True. If any directories contain
            files that are not valid datasets then an exception will
            be raised, unless the *ignore_read_error* parameter is
            True.

         external: (sequence of) `str`, optional
            Read external variables (i.e. variables which are named by
            attributes, but are not present, in the parent file given
            by the *filename* parameter) from the given external
            files. Ignored if the parent file does not contain a
            global "external_variables" attribute. Multiple external
            files may be provided, which are searched in random order
            for the required external variables.

            If an external variable is not found in any external
            files, or is found in multiple external files, then the
            relevant metadata construct is still created, but without
            any metadata or data. In this case the construct's
            `!is_external` method will return `True`.

            *Parameter example:*
              ``external='cell_measure.nc'``

            *Parameter example:*
              ``external=['cell_measure.nc']``

            *Parameter example:*
              ``external=('cell_measure_A.nc', 'cell_measure_O.nc')``

        extra: (sequence of) `str`, optional
            Create extra, independent field constructs from netCDF
            variables that correspond to particular types metadata
            constructs. The *extra* parameter may be one, or a
            sequence, of:

            ==========================  ===============================
            *extra*                     Metadata constructs
            ==========================  ===============================
            ``'field_ancillary'``       Field ancillary constructs
            ``'domain_ancillary'``      Domain ancillary constructs
            ``'dimension_coordinate'``  Dimension coordinate constructs
            ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
            ``'cell_measure'``          Cell measure constructs
            ==========================  ===============================

            This parameter replaces the deprecated *field* parameter.

            *Parameter example:*
              To create field constructs from auxiliary coordinate
              constructs: ``extra='auxiliary_coordinate'`` or
              ``extra=['auxiliary_coordinate']``.

            *Parameter example:*
              To create field constructs from domain ancillary and
              cell measure constructs: ``extra=['domain_ancillary',
              'cell_measure']``.

            An extra field construct created via the *extra* parameter
            will have a domain limited to that which can be inferred
            from the corresponding netCDF variable, but without the
            connections that are defined by the parent netCDF data
            variable. It is possible to create independent fields from
            metadata constructs that do incorporate as much of the
            parent field construct's domain as possible by using the
            `~cf.Field.convert` method of a returned field construct,
            instead of setting the *extra* parameter.

        verbose: `bool`, optional
            If True then print a description of how the contents of
            the netCDF file were parsed and mapped to CF data model
            constructs.

        warnings: `bool`, optional
            If True then print warnings when an output field construct
            is incomplete due to structural non-compliance of the
            dataset. By default such warnings are not displayed.

        ignore_read_error: `bool`, optional
            If True then ignore any file which raises an IOError
            whilst being read, as would be the case for an empty file,
            unknown file format, etc. By default the IOError is
            raised.

        fmt: `str`, optional
            Only read files of the given format, ignoring all other
            files. Valid formats are ``'NETCDF'`` for CF-netCDF files,
            ``'CFA'`` for CFA-netCDF files, ``'UM'`` for PP or UM
            fields files, and ``'CDL'`` for CDL text files. By default
            files of any of these formats are read.

        aggregate: `bool` or `dict`, optional
            If True (the default) or a dictionary (possibly empty)
            then aggregate the field constructs read in from all input
            files into as few field constructs as possible by passing
            all of the field constructs found the input files to the
            `cf.aggregate`, and returning the output of this function
            call.

            If *aggregate* is a dictionary then it is used to
            configure the aggregation process passing its contents as
            keyword arguments to the `cf.aggregate` function.

            If *aggregate* is False then the field constructs are not
            aggregated.

        squeeze: `bool`, optional
            If True then remove size 1 axes from each field construct's
            data array.

        unsqueeze: `bool`, optional
            If True then insert size 1 axes from each field
            construct's domain into its data array.

        select: (sequence of) str, optional
            Only return field constructs which have given identities,
            i.e. those that satisfy
            ``f.match_by_identity(*select)``. See
            `cf.Field.match_by_identity` for details.

        recursive: `bool`, optional
            If True then recursively read sub-directories of any
            directories specified with the *files* parameter.

        followlinks: `bool`, optional
            If True, and *recursive* is True, then also search for
            files in sub-directories which resolve to symbolic
            links. By default directories which resolve to symbolic
            links are ignored. Ignored of *recursive* is False. Files
            which are symbolic links are always followed.

            Note that setting ``recursive=True, followlinks=True`` can
            lead to infinite recursion if a symbolic link points to a
            parent directory of itself.

            This parameter replaces the deprecated *follow_symlinks*
            parameter.

        um: `dict`, optional
            For Met Office (UK) PP files and Met Office (UK) fields
            files only, provide extra decoding instructions. This
            option is ignored for input files which are notPP or
            fields files. In most cases, how to decode a file is
            inferrable from the file's contents, but if not then each
            key/value pair in the dictionary sets a decoding option as
            follows:

            ============================  =====================================
            Key                           Value
            ============================  =====================================
            ``'fmt'``                     The file format (``'PP'`` or
                                          ``'FF'``)

            ``'word_size'``               The word size in bytes (``4`` or
                                          ``8``)

            ``'endian'``                  The byte order (``'big'`` or
                                          ``'little'``)

            ``'version'``                 The UM version to be used
                                          when decoding the
                                          header. Valid versions
                                          are, for example, ``4.2``,
                                          ``'6.6.3'`` and
                                          ``'8.2'``. The default
                                          version is ``4.5``. In
                                          general, a given version
                                          is ignored if it can be
                                          inferred from the header
                                          (which is usually the case
                                          for files created by the
                                          UM at versions 5.3 and
                                          later). The exception to
                                          this is when the given
                                          version has a third
                                          element (such as the 3 in
                                          6.6.3), in which case any
                                          version in the header is
                                          ignored.

            ``'height_at_top_of_model'``  The height (in metres) of
                                          the upper bound of the top
                                          model level. By default
                                          the height at top model is
                                          taken from the top level's
                                          upper bound defined by
                                          BRSVD1 in the lookup
                                          header. If the height at
                                          top model can not be
                                          determined from the header and is
                                          not provided then no
                                          "atmosphere_hybrid_height_coordinate"
                                          dimension coordinate
                                          construct will be created.
            ============================  =====================================

            If format is specified as ``'PP'`` then the word size and
            byte order default to ``4`` and ``'big'`` respectively.

            This parameter replaces the deprecated *umversion* and
            *height_at_top_of_model* parameters.

            *Parameter example:*
              To specify that the input files are 32-bit, big-endian
              PP files: ``um={'fmt': 'PP'}``

            *Parameter example:*
              To specify that the input files are 32-bit,
              little-endian PP files from version 5.1 of the UM:
              ``um={'fmt': 'PP', 'endian': 'little', 'version': 5.1}``

            .. versionadded:: 1.5

        umversion: deprecated at version 3.0.0
            Use the *um* parameter instead.

        height_at_top_of_model: deprecated at version 3.0.0
            Use the *um* parameter instead.

        field: deprecated at version 3.0.0
            Use the *extra* parameter instead.

        follow_symlinks: deprecated at version 3.0.0
            Use the *followlinks* parameter instead.

        select_options: deprecated at version 3.0.0

    :Returns:

        `FieldList`
            The field constructs found in the input file(s). The list
            may be empty.

    **Examples:**

    >>> x = cf.read('file.nc')

    Read a file and create field constructs from CF-netCDF data
    variables as well as from the netCDF variables that correspond to
    particular types metadata constructs:

    >>> f = cf.read('file.nc', extra='domain_ancillary')
    >>> g = cf.read('file.nc', extra=['dimension_coordinate',
    ...                               'auxiliary_coordinate'])

    Read a file that contains external variables:

    >>> h = cf.read('parent.nc')
    >>> i = cf.read('parent.nc', external='external.nc')
    >>> j = cf.read('parent.nc', external=['external1.nc', 'external2.nc'])

    >>> f = cf.read('file*.nc')
    >>> f
    [<CF Field: pmsl(30, 24)>,
     <CF Field: z-squared(17, 30, 24)>,
     <CF Field: temperature(17, 30, 24)>,
     <CF Field: temperature_wind(17, 29, 24)>]

    >>> cf.read('file*.nc')[0:2]
    [<CF Field: pmsl(30, 24)>,
     <CF Field: z-squared(17, 30, 24)>]

    >>> cf.read('file*.nc')[-1]
    <CF Field: temperature_wind(17, 29, 24)>

    >>> cf.read('file*.nc', select='units=K')
    [<CF Field: temperature(17, 30, 24)>,
     <CF Field: temperature_wind(17, 29, 24)>]

    >>> cf.read('file*.nc', select='ncvar%ta')
    <CF Field: temperature(17, 30, 24)>

    '''
    if field:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            'cf.read', {'field': field}, "Use keyword 'extra' instead"
        )  # pragma: no cover

    if select_options:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            'cf.read', {'select_options': select_options}
        )  # pragma: no cover

    if follow_symlinks:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            'cf.read', {'follow_symlinks': follow_symlinks},
            "Use keyword 'followlink' instead."
        )  # pragma: no cover

    if height_at_top_of_model is not None:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            'cf.read', {'height_at_top_of_model': height_at_top_of_model},
            "Use keyword 'um' instead."
        )  # pragma: no cover

    # Parse select
    if isinstance(select, str):
        select = (select,)

    if squeeze and unsqueeze:
        raise ValueError("squeeze and unsqueeze can not both be True")

    if follow_symlinks and not recursive:
        raise ValueError(
            "Can't set follow_symlinks={0} when recursive={1}".format(
                follow_symlinks, recursive))

    # Initialize the output list of fields
    field_list = FieldList()

    if isinstance(aggregate, dict):
        aggregate_options = aggregate.copy()
        aggregate = True
    else:
        aggregate_options = {}

    aggregate_options['copy'] = False

    # Parse the extra parameter
    if extra is None:
        extra = ()
    elif isinstance(extra, str):
        extra = (extra,)

    ftypes = set()

    # Count the number of fields (in all files) and the number of
    # files
    field_counter = -1
    file_counter = 0

    for file_glob in flat(files):
        # Expand variables
        file_glob = os.path.expanduser(os.path.expandvars(file_glob))

        if file_glob.startswith('http://'):
            # Do not glob a URL
            files2 = (file_glob,)
        else:
            # Glob files on disk
            files2 = glob(file_glob)

            if not files2 and not ignore_read_error:
                open(file_glob, 'rb')

            files3 = []
            for x in files2:
                if isdir(x):
                    # Walk through directories, possibly recursively
                    for path, subdirs, filenames in os.walk(
                            x, followlinks=followlinks):
                        files3.extend(os.path.join(path, f) for f in filenames)
                        if not recursive:
                            break
                else:
                    files3.append(x)
            # --- End: for

            files2 = files3

        for filename in files2:
            if verbose:
                print('File: {0}'.format(filename))  # pragma: no cover

            if um:
                ftype = 'UM'
            else:
                try:
                    ftype = file_type(filename)
                except Exception as error:
                    if not ignore_read_error:
                        raise Exception(error)

                    if verbose:
                        print('WARNING: {}'.format(error))  # pragma: no cover

                    continue
            # --- End: if

            ftypes.add(ftype)

            # --------------------------------------------------------
            # Read the file into fields
            # --------------------------------------------------------
            fields = _read_a_file(
                filename, ftype=ftype,
                external=external,
                ignore_read_error=ignore_read_error,
                verbose=verbose, warnings=warnings,
                aggregate=aggregate,
                aggregate_options=aggregate_options,
                selected_fmt=fmt, um=um,
                extra=extra,
                height_at_top_of_model=height_at_top_of_model,
                chunk=chunk
            )

            # --------------------------------------------------------
            # Select matching fields (not from UM files)
            # --------------------------------------------------------
            if select and ftype != 'UM':
                fields = fields.select_by_identity(*select)

            # --------------------------------------------------------
            # Add this file's fields to those already read from other
            # files
            # --------------------------------------------------------
            field_list.extend(fields)

            field_counter = len(field_list)
            file_counter += 1
        # --- End: for
    # --- End: for

    # Print some informative messages
    if verbose:
        print(
            "Read {0} field{1} from {2} file{3}".format(
                field_counter, _plural(field_counter), file_counter,
                _plural(file_counter)
            )
        )  # pragma: no cover

    # ----------------------------------------------------------------
    # Aggregate the output fields
    # ----------------------------------------------------------------
    if aggregate and len(field_list) > 1:
        if verbose:
            org_len = len(field_list)  # pragma: no cover

        field_list = cf_aggregate(field_list, **aggregate_options)

        if verbose:
            n = len(field_list)  # pragma: no cover
            print('{0} input field{1} aggregated into {2} field{3}'.format(
                org_len, _plural(org_len),
                n, _plural(n)))  # pragma: no cover
    # --- End: if

    # ----------------------------------------------------------------
    # Sort by netCDF variable name
    # ----------------------------------------------------------------
    if len(field_list) > 1:
        field_list.sort(key=lambda f: f.nc_get_variable(''))

    # ----------------------------------------------------------------
    # Add standard names to UM/PP fields (post aggregation)
    # ----------------------------------------------------------------
    for f in field_list:
        standard_name = f._custom.get('standard_name', None)
        if standard_name is not None:
            f.set_property('standard_name', standard_name)
            del f._custom['standard_name']
    # --- End: for

    # ----------------------------------------------------------------
    # Select matching fields from UM/PP fields (post setting of
    # standard names)
    # ----------------------------------------------------------------
    if select and 'UM' in ftypes:
        field_list = field_list.select_by_identity(*select)

    # ----------------------------------------------------------------
    # Squeeze size one dimensions from the data arrays. Do one of:
    #
    # 1) Squeeze the fields, i.e. remove all size one dimensions from
    #    all field data arrays
    #
    # 2) Unsqueeze the fields, i.e. Include all size 1 domain
    #    dimensions in the data array.
    #
    # 3) Nothing
    # ----------------------------------------------------------------
    if squeeze:
        for f in field_list:
            f.squeeze(inplace=True)
    elif unsqueeze:
        for f in field_list:
            f.unsqueeze(inplace=True)
    # --- End: if

    if nfields is not None and len(field_list) != nfields:
        raise ValueError(
            "{} field{} requested but {} fields found in file{}".format(
                nfields, _plural(nfields), len(field_list),
                _plural(file_counter)
            )
        )

    return field_list


def _plural(n):  # pragma: no cover
    '''Return a suffix which reflects a word's plural.

    '''
    return 's' if n != 1 else ''  # pragma: no cover


def _read_a_file(filename, ftype=None, aggregate=True,
                 aggregate_options=None, ignore_read_error=False,
                 verbose=False, warnings=False, external=None,
                 selected_fmt=None, um=None, extra=None,
                 height_at_top_of_model=None, chunk=True):
    '''Read the contents of a single file into a field list.

    :Parameters:

        filename: `str`
            The file name.

        ftype: `str`
            TODO

        aggregate_options: `dict`, optional
            The keys and values of this dictionary may be passed as
            keyword parameters to an external call of the aggregate
            function.

        ignore_read_error: `bool`, optional
            If True then return an empty field list if reading the
            file produces an IOError, as would be the case for an
            empty file, unknown file format, etc. By default the
            IOError is raised.

        verbose: `bool`, optional
            If True then print information to stdout.

    :Returns:

        `FieldList`
            The fields in the file.

    '''
    if aggregate_options is None:
        aggregate_options = {}
    # Find this file's type
    fmt = None
    word_size = None
    endian = None
    height_at_top_of_model = None
    umversion = 405

    if um:
        # ftype = 'UM'
        fmt = um.get('fmt')
        word_size = um.get('word_size')
        endian = um.get('endian')
        umversion = um.get('version')
        height_at_top_of_model = um.get('height_at_top_of_model')
        if fmt in ('PP', 'pp', 'pP', 'Pp'):
            fmt = fmt.upper()
            # For PP format, there is a default word size and
            # endian-ness
            if word_size is None:
                word_size = 4
            if endian is None:
                endian = 'big'
        # --- End: if

        if umversion is not None:
            umversion = float(str(umversion).replace('.', '0', 1))
#    else:
#        try:
#            ftype = file_type(filename)
#        except Exception as error:
#            if not ignore_read_error:
#                raise Exception(error)
#
#            if verbose:
#                print('WARNING: {}'.format(error))  # pragma: no cover
#
#            return FieldList()
    # --- End: if

    extra_read_vars = {
        'chunk': chunk,
        'fmt': selected_fmt,
        'ignore_read_error': ignore_read_error,
        # 'cfa' defaults to False. If the file has
        # "CFA" in its Conventions global attribute
        # then 'cfa' will be changed to True in
        # netcdf.read
        'cfa': False,
    }

    # ----------------------------------------------------------------
    # Still here? Read the file into fields.
    # ----------------------------------------------------------------
    if ftype == 'CDL':
        # Create a temporary netCDF file from input CDL
        ftype = 'netCDF'
        cdl_filename = filename
        filename = netcdf.cdl_to_netcdf(filename)
        extra_read_vars['fmt'] = 'NETCDF'

        if not netcdf.is_netcdf_file(filename):
            if ignore_read_error:
                if verbose:
                    print(
                        "WARNING: Can't determine format of file {} generated "
                        "from CDL file {}".format(filename, cdl_filename)
                    )  # pragma: no cover

                return FieldList()
            else:
                raise IOError(
                    "Can't determine format of file {} generated from CDL "
                    "file {}".format(filename, cdl_filename)
                )
    # --- End: if

    if ftype == 'netCDF' and extra_read_vars['fmt'] in (None, 'NETCDF', 'CFA'):
        fields = netcdf.read(filename, external=external, extra=extra,
                             verbose=verbose, warnings=warnings,
                             extra_read_vars=extra_read_vars)

    elif ftype == 'UM' and extra_read_vars['fmt'] in (None, 'UM'):
        fields = UM.read(filename, um_version=umversion,
                         verbose=verbose, set_standard_name=False,
                         height_at_top_of_model=height_at_top_of_model,
                         fmt=fmt, word_size=word_size, endian=endian,
                         chunk=chunk)

        # PP fields are aggregated intrafile prior to interfile
        # aggregation
        if aggregate:
            # For PP fields, the default is strict_units=False
            if 'strict_units' not in aggregate_options:
                aggregate_options['relaxed_units'] = True
    else:
        fields = ()

    # ----------------------------------------------------------------
    # Check for cyclic dimensions
    # ----------------------------------------------------------------
    for f in fields:
        f.autocyclic()

    # ----------------------------------------------------------------
    # Return the fields
    # ----------------------------------------------------------------
    return FieldList(fields)


def file_type(filename):
    '''Return the file format.

    :Parameters:

        filename: `str`
            The file name.

    :Returns:

        `str`
            The format type of the file. One of ``'netCDF'``, ``'UM'``
            or ``'CDL'``.

    **Examples:**

    >>> file_type(filename)
    'netCDF'

    '''
    # ----------------------------------------------------------------
    # NetCDF
    # ----------------------------------------------------------------
    if netcdf.is_netcdf_file(filename):
        return 'netCDF'

    # ----------------------------------------------------------------
    # PP or FF
    # ----------------------------------------------------------------
    if UM.is_um_file(filename):
        return 'UM'

    # ----------------------------------------------------------------
    # CDL
    # ----------------------------------------------------------------
    if netcdf.is_cdl_file(filename):
        return 'CDL'

    # Still here?
    raise IOError("Can't determine format of file {}".format(filename))



import json

from ast import literal_eval as ast_literal_eval

import cfdm

from ...constants import _file_to_fh


class NetCDFRead(cfdm.read_write.netcdf.NetCDFRead):
    '''TODO

    .. versionadded:: 3.0.0

    '''
    def _ncdimensions(self, ncvar):
        '''Return a list of the netCDF dimensions corresponding to a netCDF
    variable.

    If the variable has been compressed then the *implied
    uncompressed* dimensions are returned.

    For a CFA variable, the netCDF dimensions are taken from the
    'cfa_dimensions' netCDF attribute.

    .. versionadded:: 3.0.0

    :Parameters:

        ncvar: `str`
            The netCDF variable name.

    :Returns:

        `list`
            The netCDF dimension names spanned by the netCDF variable.

    **Examples:**

    >>> n._ncdimensions('humidity')
    ['time', 'lat', 'lon']

        '''
        g = self.read_vars

        cfa = (g['cfa'] and
               ncvar not in g['external_variables'] and
               g['variable_attributes'][ncvar].get('cf_role') ==
               'cfa_variable')

        if not cfa:
            return super()._ncdimensions(ncvar)

        # Still here?
        ncdimensions = g['variable_attributes'][ncvar].get(
            'cfa_dimensions', '').split()

        return list(map(str, ncdimensions))

    def _get_domain_axes(self, ncvar, allow_external=False):
        '''Return the domain axis identifiers that correspond to a netCDF
    variable's netCDF dimensions.

    For a CFA variable, the netCDF dimensions are taken from the
    'cfa_dimensions' netCDF attribute.

    :Parameter:

        ncvar: `str`
            The netCDF variable name.

        allow_external: `bool`
            If `True` and *ncvar* is an external variable then return an
            empty list.

    :Returns:

        `list`

    **Examples:**

    >>> r._get_domain_axes('areacello')
    ['domainaxis0', 'domainaxis1']

    >>> r._get_domain_axes('areacello', allow_external=True)
    []

        '''
        g = self.read_vars

        cfa = (g['cfa'] and
               ncvar not in g['external_variables'] and
               g['variable_attributes'][ncvar].get('cf_role') ==
               'cfa_variable')

        if not cfa:
            return super()._get_domain_axes(ncvar=ncvar,
                                            allow_external=allow_external)

        # Still here?
        cfa_dimensions = g['variable_attributes'][ncvar].get(
            'cfa_dimensions', '').split()

        ncdim_to_axis = g['ncdim_to_axis']
        axes = [ncdim_to_axis[ncdim] for ncdim in cfa_dimensions
                if ncdim in ncdim_to_axis]

        return axes

    def _create_data(self, ncvar, construct=None,
                     unpacked_dtype=False, uncompress_override=None,
                     parent_ncvar=None):
        '''TODO

    .. versionadded:: 3.0.0

    :Parameters:


    :Returns:

        `Data`

        '''
        g = self.read_vars

        is_cfa_variable = (
            g['cfa'] and
            construct.get_property('cf_role', None) == 'cfa_variable')

        if not is_cfa_variable:
            # --------------------------------------------------------
            # Create data for a normal netCDF variable
            # --------------------------------------------------------
            return super()._create_data(
                ncvar=ncvar,
                construct=construct,
                unpacked_dtype=unpacked_dtype,
                uncompress_override=uncompress_override,
                parent_ncvar=parent_ncvar
            )

        # ------------------------------------------------------------
        # Still here? Then create data for a CFA netCDF variable
        # ------------------------------------------------------------
#        print ('    Creating data from CFA variable', repr(ncvar),
#               repr(construct))
        try:
            cfa_data = json.loads(construct.get_property('cfa_array'))
        except ValueError as error:
            raise ValueError(
                "Error during JSON-decoding of netCDF attribute 'cfa_array': "
                "{}".format(error)
            )

        variable = g['variables'][ncvar]

        cfa_data['file'] = g['filename']
        cfa_data['Units'] = construct.Units
        cfa_data['fill_value'] = construct.fill_value()
        cfa_data['_pmshape'] = cfa_data.pop('pmshape', ())
        cfa_data['_pmaxes'] = cfa_data.pop('pmdimensions', ())

        base = cfa_data.get('base', None)
        if base is not None:
            cfa_data['base'] = pathjoin(dirname(g['filename']), base)

        ncdimensions = construct.get_property('cfa_dimensions', '').split()
        dtype = variable.dtype

        if dtype is str:
            # netCDF string types have a dtype of `str`, which needs
            # to be reset as a numpy.dtype, but we don't know what
            # without reading the data, so set it to None for now.
            dtype = None

        # UNICODE???? TODO
        if self._is_char(ncvar) and dtype.kind in 'SU' and ncdimensions:
            # strlen = len(nc.dimensions[ncdimensions[-1]])
            strlen = g['nc'].dimensions[ncdimensions[-1]].size
            if strlen > 1:
                ncdimensions.pop()
                dtype = numpy_dtype('S{0}'.format(strlen))
        # --- End: if

        cfa_data['dtype'] = dtype
        cfa_data['_axes'] = ncdimensions
#        cfa_data['shape'] = [len(nc.dimensions[ncdim])
#                             for ncdim in ncdimensions]
        cfa_data['shape'] = [g['nc'].dimensions[ncdim].size
                             for ncdim in ncdimensions]

        for attrs in cfa_data['Partitions']:

            # FORMAT
            sformat = attrs.get('subarray', {}).pop('format', 'netCDF')
            if sformat is not None:
                attrs['format'] = sformat

            # DTYPE
            dtype = attrs.get('subarray', {}).pop('dtype', None)
            if dtype not in (None, 'char'):
                attrs['subarray']['dtype'] = numpy_dtype(dtype)

            # UNITS and CALENDAR
            units = attrs.pop('punits', None)
            calendar = attrs.pop('pcalendar', None)
            if units is not None or calendar is not None:
                attrs['Units'] = Units(units, calendar)

            # AXES
            pdimensions = attrs.pop('pdimensions', None)
            if pdimensions is not None:
                attrs['axes'] = pdimensions

            # REVERSE
            reverse = attrs.pop('reverse', None)
            if reverse is not None:
                attrs['reverse'] = reverse

            # LOCATION: Change to python indexing (i.e. range does not
            #           include the final index)
            for r in attrs['location']:
                r[1] += 1

            # PART: Change to python indexing (i.e. slice range does
            #       not include the final index)
            part = attrs.get('part', None)
            if part:
                p = []
                for x in ast_literal_eval(part):
                    if isinstance(x, list):
                        if x[2] > 0:
                            p.append(slice(x[0], x[1]+1, x[2]))
                        elif x[1] == 0:
                            p.append(slice(x[0], None, x[2]))
                        else:
                            p.append(slice(x[0], x[1]-1, x[2]))
                    else:
                        p.append(list(x))
                # --- End: for

                attrs['part'] = p
        # --- End: for

        construct.del_property('cf_role')
        construct.del_property('cfa_array')
        construct.del_property('cfa_dimensions', None)

        out = self._create_Data(loadd=cfa_data)

        return out

    def _create_Data(self, array=None, units=None, calendar=None,
                     ncvar=None, loadd=None, **kwargs):
        '''TODO

    .. versionadded:: 3.0.0

        '''
        try:
            compressed = array.get_compression_type()  # TODO
        except AttributeError:
            compressed = False

        if not compressed:
            # Do not chunk compressed data (for now ...)
            chunk = False
        else:
            chunk = self.read_vars.get('chunk', True)

        return super()._create_Data(array=array, units=units,
                                    calendar=calendar, ncvar=ncvar,
                                    loadd=loadd, **kwargs)

    def _customize_read_vars(self):
        '''TODO

    .. versionadded:: 3.0.0

        '''
        super()._customize_read_vars()

        g = self.read_vars

        # ------------------------------------------------------------
        # Find out if this is a CFA file
        # ------------------------------------------------------------
        g['cfa'] = 'CFA' in g['global_attributes'].get(
            'Conventions', ())  # TODO

        # ------------------------------------------------------------
        # Do not create fields from CFA private variables
        # ------------------------------------------------------------
        if g['cfa']:
            for ncvar in g['variables']:
                if (g['variable_attributes'][ncvar].get('cf_role', None) ==
                        'cfa_private'):
                    g['do_not_create_field'].add(ncvar)
        # --- End: if

        # ------------------------------------------------------------
        #
        # ------------------------------------------------------------
        if g['cfa']:
            for ncvar, ncdims in tuple(g['variable_dimensions'].items()):
                if ncdims != ():
                    continue

                if not (ncvar not in g['external_variables'] and
                        g['variable_attributes'][ncvar].get('cf_role') ==
                        'cfa_variable'):
                    continue

                ncdimensions = g['variable_attributes'][ncvar].get(
                    'cfa_dimensions', '').split()
                if ncdimensions:
                    g['variable_dimensions'][ncvar] = tuple(
                        map(str, ncdimensions))

    def file_open(self, filename):
        '''Open the netCDf file for reading.

    :Paramters:

        filename: `str`
            The netCDF file to be read.

    :Returns:

        `netCDF4.Dataset`
            The object for the file.

        '''
        out = super().file_open(filename)
        _file_to_fh['netCDF'].pop(filename, None)
        return out

# --- End: class



import cfdm

from .query import Query


class Constructs(cfdm.Constructs):
    '''A container for metadata constructs.

    Calling a `Constructs` instance selects metadata constructs by
    identity and is an alias for the `filter_by_identity` method.

    **Examples:**

    Select constructs that have a "standard_name" property of 'latitude':

    >>> d = c('latitude')

    .. versionadded:: 3.0.0

    '''

    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return super().__repr__().replace('<', '<CF ', 1)

    def _matching_values(self, value0, construct, value1):
        '''TODO
        '''
        if isinstance(value0, Query):
            return value0.evaluate(value1)

        return super()._matching_values(value0, construct, value1)

#     def domain_axis_key(self, identity, default=ValueError()):
#         '''Return the key of the domain axis construct that is spanned by 1-d
# coordinate constructs.
#
# .. versionadded:: 3.0.0
#
# :Parameters:
#
#     identity:
#
#         Select the 1-d coordinate constructs that have the given
#         identity.
#
#         An identity is specified by a string (e.g. ``'latitude'``,
#         ``'long_name=time'``, etc.); or a compiled regular expression
#         (e.g. ``re.compile('^atmosphere')``), for which all constructs
#         whose identities match (via `re.search`) are selected.
#
#         Each coordinate construct has a number of identities, and is
#         selected if any of them match any of those provided. A
#         construct's identities are those returned by its `!identities`
#         method. In the following example, the construct ``x`` has four
#         identities:
#
#            >>> x.identities()
#            ['time', 'long_name=Time', 'foo=bar', 'ncvar%T']
#
#         In addition, each construct also has an identity based its
#         construct key (e.g. ``'key%dimensioncoordinate2'``)
#
#         Note that in the output of a `print` call or `!dump` method, a
#         construct is always described by one of its identities, and so
#         this description may always be used as an *identity* argument.
#
#     default: optional
#         Return the value of the *default* parameter if a domain axis
#         construct can not be found. If set to an `Exception` instance
#         then it will be raised instead.
#
# :Returns:
#
#     `str`
#         The key of the domain axis construct that is spanned by the
#         data of the selected 1-d coordinate constructs.
#
# **Examples:**
#
# TODO
#
#         '''
#         # Try for index
#         try:
#             da_key = self.get_data_axes(default=None)[identity]
#         except TypeError:
#             pass
#         except IndexError:
#             return self._default(
#                 default,
#                 "Index does not exist for field construct data dimenions")
#         else:
#             identity = da_key
#
#         domain_axes = self.domain_axes(identity)
#         if len(domain_axes) == 1:
#             # identity is a unique domain axis construct identity
#             da_key = domain_axes.key()
#         else:
#             # identity is not a unique domain axis construct identity
#             da_key = self.domain_axis_key(identity, default=default)
#
#         if key:
#             return da_key
#
#         return self.constructs[da_key]

    def filter_by_identity(self, *identities):
        '''Select metadata constructs by identity.

    .. versionadded:: 3.0.0

    .. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
                 `filter_by_measure`, `filter_by_method`,
                 `filter_by_naxes`, `filter_by_ncdim`,
                 `filter_by_ncvar`, `filter_by_property`,
                 `filter_by_size`, `filter_by_type`,
                 `filters_applied`, `inverse_filter`, `unfilter`

    :Parameters:

        identities: optional

            Select constructs that have any of the given identities or
            construct keys.

            An identity is specified by a string (e.g. ``'latitude'``,
            ``'long_name=time'``, etc.); or a compiled regular
            expression (e.g. ``re.compile('^atmosphere')``), for which
            all constructs whose identities match (via `re.search`)
            are selected.

            If no identities are provided then all constructs are selected.

            Each construct has a number of identities, and is selected
            if any of them match any of those provided. A construct's
            identities are those returned by its `!identities`
            method. In the following example, the construct ``x`` has
            five identities:

               >>> x.identities()
               ['time', 'long_name=Time', 'foo=bar', 'T', 'ncvar%t']

            A construct key may optionally have the ``'key%'``
            prefix. For example ``'dimensioncoordinate2'`` and
            ``'key%dimensioncoordinate2'`` are both acceptable keys.

            Note that the identifiers of a metadata construct in the
            output of a `print` or `!dump` call are always one of its
            identities, and so may always be used as an *identities*
            argument.

            Domain axis constructs may also be identified by their
            position in the field construct's data array. Positions
            are specified by either integers.

            .. note:: This is an extension to the functionality of
                      `cfdm.Constucts.filter_by_identity`.

    :Returns:

        `Constructs`
            The selected constructs and their construct keys.

    **Examples:**

    Select constructs that have a "standard_name" property of
    'latitude':

    >>> d = c.filter_by_identity('latitude')

    Select constructs that have a "long_name" property of 'Height':

    >>> d = c.filter_by_identity('long_name=Height')

    Select constructs that have a "standard_name" property of
    'latitude' or a "foo" property of 'bar':

    >>> d = c.filter_by_identity('latitude', 'foo=bar')

    Select constructs that have a netCDF variable name of 'time':

    >>> d = c.filter_by_identity('ncvar%time')

        '''
#        field_data_axes = self._field_data_axes
#
#        if field_data_axes is not None:
#            # Allows integer data domain axis positions, do we want this? TODO
#            new_identities = []
#            for i in identities:
#                try:
#                    _ = field_data_axes[i]
#                except IndexError:
#                    new_identities.append(i)
#                else:
#                    if isinstance(_, str):
#                        new_identities.append('key%'+_)
#                    else:
#                        new_identities.extend(['key%'+axis for axis in _])
#        else:
#            new_identities = identities
#

        # Allow keys without the 'key%' prefix
        identities = list(identities)
        for n, identity in enumerate(identities):
            if identity in self:
                identities[n] = 'key%'+identity
        # --- End: for

        return super().filter_by_identity(*identities)


# --- End: class



import datetime
import os
import time
import unittest

import numpy

import cf

class DimensionCoordinateTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')

        dim1 = cf.DimensionCoordinate()
        dim1.standard_name = 'latitude'
        a = numpy.array([-30, -23.5, -17.8123, -11.3345, -0.7, -0.2, 0, 0.2, 0.7, 11.30003, 17.8678678, 23.5, 30])
        dim1.set_data(cf.Data(a, 'degrees_north'))
        bounds = cf.Bounds()
        b = numpy.empty(a.shape + (2,))
        b[:, 0] = a - 0.1
        b[:, 1] = a + 0.1
        bounds.set_data(cf.Data(b))
        dim1.set_bounds(bounds)
        self.dim = dim1

    def test_DimensionCoordinate__repr__str__dump(self):
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()

        _ = repr(x)
        _ = str(x)
        _ = x.dump(display=False)

        self.assertTrue(x.isdimension)


    def test_DimensionCoordinate_convert_reference_time(self):
        d = cf.DimensionCoordinate()
        d.set_data(cf.Data([1, 2, 3], 'months since 2004-1-1', calendar='gregorian'))
        self.assertTrue((d.array == [1., 2, 3]).all())

        e = d.copy()
        self.assertTrue(e.convert_reference_time(calendar_months=True, inplace=True) is None)

        f = d.convert_reference_time(calendar_months=True)

        for x in (e, f):
            self.assertTrue((x.array == [31., 60., 91.]).all())
            self.assertTrue((x.datetime_array == [cf.dt('2004-02-01 00:00:00', calendar='gregorian'),
                                                  cf.dt('2004-03-01 00:00:00', calendar='gregorian'),
                                                  cf.dt('2004-04-01 00:00:00', calendar='gregorian')]).all())

        self.assertTrue((d.array == [1., 2, 3]).all())


        d = cf.DimensionCoordinate()
        d.set_data(cf.Data([1, 2, 3], 'months since 2004-1-1', calendar='360_day'))
        e = d.copy()
        self.assertTrue(e.convert_reference_time(calendar_months=True, inplace=True) is None)

        f = d.convert_reference_time(calendar_months=True)

        for x in (e, f):
            self.assertTrue((x.array == [30., 60., 90.]).all())
            self.assertTrue((x.datetime_array == [cf.dt('2004-02-01 00:00:00', calendar='360_day'),
                                                  cf.dt('2004-03-01 00:00:00', calendar='360_day'),
                                                  cf.dt('2004-04-01 00:00:00', calendar='360_day')]).all())

        self.assertTrue((d.array == [1., 2, 3]).all())


        d = cf.DimensionCoordinate()
        d.set_data(cf.Data([1, 2, 3], 'months since 2004-1-1', calendar='noleap'))
        e = d.copy()
        self.assertTrue(e.convert_reference_time(calendar_months=True, inplace=True) is None)

        f = d.convert_reference_time(calendar_months=True)

        for x in (e, f):
            self.assertTrue((x.array == [31., 59., 90.]).all())
            self.assertTrue((x.datetime_array == [cf.dt('2004-02-01 00:00:00', calendar='noleap'),
                                                  cf.dt('2004-03-01 00:00:00', calendar='noleap'),
                                                  cf.dt('2004-04-01 00:00:00', calendar='noleap')]).all())

        self.assertTrue((d.array == [1., 2, 3]).all())


    def test_DimensionCoordinate_roll(self):
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()

        with self.assertRaises(Exception):
            x.roll(0, 3)

        x.period(360)
        _ = x.roll(0, 3)
        _ = x.roll(-1, 3)
        with self.assertRaises(Exception):
            _ = x.roll(2, 3)

        a = x[0]
        _ = a.roll(0, 3)
        self.assertTrue(a.roll(0, 3, inplace=True) is None)

        _ = x.roll(0, 0)
        _ = x.roll(0, 3, inplace=True)
        self.assertTrue(x.roll(0, 0, inplace=True) is None)

        _ = x._centre(360)
        _ = x.flip()._centre(360)

        # Test roll on coordinate without bounds:
        g = f.copy()
        g.dimension_coordinate('X').del_bounds()

        for shift_by in [1, -1, g.shape[2]]:  # vary roll direction and extent
            g_rolled = g.roll('X', shift=shift_by)

            if shift_by == g.shape[2]:  # shift_by equal to the roll axis size
                g_rolled_0 = g.roll('X', shift=0)
                # A roll of the axes size, or 0, should not change the array:
                self.assertTrue((g_rolled.array == g.array).all())
                self.assertTrue((g_rolled.array == g_rolled_0.array).all())

            for index in range(0, 10):  # check all elements are rolled
                self.assertTrue(
                    g_rolled.array[0, index, 0] == g.array[0, index, -shift_by]
                )

    def test_DimensionCoordinate_cellsize(self):
        d = self.dim.copy()

        c = d.cellsize
        self.assertTrue(numpy.allclose(c.array, 0.2))

        self.assertTrue(d.Units.equals(cf.Units('degrees_north')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('degrees_north')))

        d.override_units('km', inplace=True)
        self.assertTrue(d.Units.equals(cf.Units('km')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('km')))

        c = d.cellsize
        self.assertTrue(c.Units.equals(cf.Units('km')))

        d.del_bounds()
        c = d.cellsize
        self.assertTrue(numpy.allclose(c.array, 0))


    def test_DimensionCoordinate_override_units(self):
        d = self.dim.copy()

        self.assertTrue(d.Units.equals(cf.Units('degrees_north')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('degrees_north')))

        d.override_units('km', inplace=True)
        self.assertTrue(d.Units.equals(cf.Units('km')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('km')))

        c = d.cellsize
        self.assertTrue(c.Units.equals(cf.Units('km')))


    def test_DimensionCoordinate_override_calendar(self):
        d = self.dim.copy()

        self.assertTrue(d.Units.equals(cf.Units('degrees_north')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('degrees_north')))

        d.override_units('days since 2000-01-01', inplace=True)
        self.assertTrue(d.Units.equals(cf.Units('days since 2000-01-01')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('days since 2000-01-01')))

        d.override_calendar('360_day', inplace=True)
        self.assertTrue(d.Units.equals(cf.Units('days since 2000-01-01', calendar='360_day')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('days since 2000-01-01', calendar='360_day')))

        d.override_calendar('365_day', inplace=True)
        self.assertTrue(d.Units.equals(cf.Units('days since 2000-01-01', calendar='365_day')))
        self.assertTrue(d.bounds.Units.equals(cf.Units('days since 2000-01-01', calendar='365_day')))


    def test_DimensionCoordinate_bounds(self):
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()

        _ = x.upper_bounds
        _ = x.lower_bounds

        self.assertTrue(x.increasing)

        y = x.flip()
        self.assertTrue(y.decreasing)
        self.assertTrue(y.upper_bounds.equals(x.upper_bounds[::-1]))
        self.assertTrue(y.lower_bounds.equals(x.lower_bounds[::-1]))

        c = x.cellsize
        c = y.cellsize

        y.del_bounds()

        b = y.create_bounds()


    def test_DimensionCoordinate_properties(self):
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()

        x.positive = 'up'
        self.assertTrue(x.positive == 'up')
        del x.positive

        x.axis = 'Z'
        self.assertTrue(x.axis == 'Z')
        del x.axis

        x.axis = 'T'
        self.assertTrue(x.ndim == 1)


    def test_DimensionCoordinate_insert_dimension(self):
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()

        self.assertTrue(x.shape == (9,))
        self.assertTrue(x.bounds.shape == (9, 2))

        y = x.insert_dimension(0)
        self.assertTrue(y.shape == (1, 9))
        self.assertTrue(y.bounds.shape == (1, 9, 2), y.bounds.shape)

        x.insert_dimension(-1, inplace=True)
        self.assertTrue(x.shape == (9, 1))
        self.assertTrue(x.bounds.shape == (9, 1, 2), x.bounds.shape)


    def test_DimensionCoordinate_binary_operation(self):
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()

        d = x.array
        b = x.bounds.array
        d2= numpy.expand_dims(d, -1)

        # --------------------------------------------------------
        # Out-of-place addition
        # --------------------------------------------------------
        c = x + 2
        self.assertTrue((c.array == d + 2).all())
        self.assertTrue((c.bounds.array == b + 2).all())

        c = x + x
        self.assertTrue((c.array == d + d).all())
        self.assertTrue((c.bounds.array == b + d2).all())

        c = x + 2
        self.assertTrue((c.array == d + 2).all())
        self.assertTrue((c.bounds.array == b + 2).all())

        self.assertTrue((x.array == d).all())
        self.assertTrue((x.bounds.array == b).all())

        # --------------------------------------------------------
        # In-place addition
        # --------------------------------------------------------
        x += 2
        self.assertTrue((x.array == d + 2).all())
        self.assertTrue((x.bounds.array == b + 2).all())

        x += x
        self.assertTrue((x.array == (d+2) * 2).all())
        self.assertTrue((x.bounds.array == b+2 + d2+2).all())

        x += 2
        self.assertTrue((x.array == (d+2)*2 + 2).all())
        self.assertTrue((x.bounds.array == b+2 + d2+2 + 2).all())

        # --------------------------------------------------------
        # Out-of-place addition (no bounds)
        # --------------------------------------------------------
        f = cf.read(self.filename)[0]
        x = f.dimension_coordinates('X').value()
        x.del_bounds()

        self.assertFalse(x.has_bounds())

        d = x.array

        c = x + 2
        self.assertTrue((c.array == d + 2).all())

        c = x + x
        self.assertTrue((c.array == d + d).all())

        c = x + 2
        self.assertTrue((c.array == d + 2).all())

        self.assertTrue((x.array == d).all())

        # --------------------------------------------------------
        # In-place addition (no bounds)
        # --------------------------------------------------------
        x += 2
        self.assertTrue((x.array == d + 2).all())

        x += x
        self.assertTrue((x.array == (d+2) * 2).all())

        x += 2
        self.assertTrue((x.array == (d+2)*2 + 2).all())


# --- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)
